﻿// See "Firefly Algorithms for Multimodal Optimization",
// X. Yang, 5th Intenational Symposium on Stochastic Algorithms, 2009.

using System;
namespace FireflyAlgorithm
{
  class FireflyProgram
  {
    static void Main(string[] args)
    {
      Console.WriteLine("\nBegin firefly algorithm optimization demo\n");
      Console.WriteLine("Goal is to solve the Michalewicz benchmark function");
      Console.WriteLine("The function has a known minimum value of -4.687658");
      Console.WriteLine("x = 2.2029 1.5707 1.2850 1.9231 1.7205");

      int numFireflies = 40; // typically 15-40
      int dim = 5;
      int maxEpochs = 1000;
      int seed = 0;
      //double errThresh = 1.0E-5; // early exit

      Console.WriteLine("\nSetting numFireflies        = " + numFireflies);
      Console.WriteLine("Setting problem dim         = " + dim);
      Console.WriteLine("Setting maxEpochs           = " + maxEpochs);
      Console.WriteLine("Setting initialization seed = " + seed);
      //Console.WriteLine("Setting early-exit error    = " + errThresh.ToString("F5"));

      Console.WriteLine("\nStarting firefly algorithm\n");
      double[] bestPosition = Solve(numFireflies, dim, seed, maxEpochs);
      Console.WriteLine("\nFinished\n");

      Console.WriteLine("Best solution found: ");
      Console.Write("x = ");
      ShowVector(bestPosition, 4, true);
      double z = Michalewicz(bestPosition);
      Console.Write("Value of function at best position = ");
      Console.WriteLine(z.ToString("F6"));
      double error = Error(bestPosition);
      Console.Write("Error at best position = ");
      Console.WriteLine(error.ToString("F4"));

      Console.WriteLine("\nEnd firefly demo\n");
      Console.ReadLine();
    } // Main

    static void ShowVector(double[] v, int dec, bool nl)
    {
      for (int i = 0; i < v.Length; ++i)
        Console.Write(v[i].ToString("F" + dec) + " ");
      if (nl == true)
        Console.WriteLine("");
    }

    static double[] Solve(int numFireflies, int dim, int seed, int maxEpochs)
    {
      Random rnd = new Random(seed);
      double minX = 0.0; // specific to Michalewicz function
      double maxX = 3.2;

      double B0 = 1.0;  // beta (attractiveness base)
      //double betaMin = 0.20;
      double g = 1.0;   // gamma (absorption for attraction)
      double a = 0.20;    // alpha
      //double a0 = 1.0;    // base alpha for decay
      int displayInterval = maxEpochs / 10;
 
      double bestError = double.MaxValue;
      double[] bestPosition = new double[dim]; // best ever

      Firefly[] swarm = new Firefly[numFireflies]; // all null

      // initialize swarm at random positions
      for (int i = 0; i < numFireflies; ++i)
      {
        swarm[i] = new Firefly(dim); // position 0, error and intensity 0.0
        for (int k = 0; k < dim; ++k) // random position
          swarm[i].position[k] = (maxX - minX) * rnd.NextDouble() + minX;
        swarm[i].error = Error(swarm[i].position); // associated error
        swarm[i].intensity = 1 / (swarm[i].error + 1); // +1 prevent div by 0
        if (swarm[i].error < bestError)
        {
          bestError = swarm[i].error;
          for (int k = 0; k < dim; ++k)
            bestPosition[k] = swarm[i].position[k];
        }
      }

      int epoch = 0;
      while (epoch < maxEpochs) // main processing
      {
        //if (bestError < errThresh) break; // are we good?
        if (epoch % displayInterval == 0 && epoch < maxEpochs) // show progress?
        {
          string sEpoch = epoch.ToString().PadLeft(6);
          Console.Write("epoch = " + sEpoch);
          Console.WriteLine("   error = " + bestError.ToString("F14"));
        }

        for (int i = 0; i < numFireflies; ++i) // each firefly
        {
          for (int j = 0; j < numFireflies; ++j) // each other firefly. weird!
          {
            if (swarm[i].intensity < swarm[j].intensity)
            {
              // curr firefly i is less intense (i is worse) so move i toward j
              double r = Distance(swarm[i].position, swarm[j].position);
              double beta = B0 * Math.Exp(-g * r * r); // original 
              //double beta = (B0 - betaMin) * Math.Exp(-g * r * r) + betaMin; // better
              //double a = a0 * Math.Pow(0.98, epoch); // better
              for (int k = 0; k < dim; ++k)
              {
                swarm[i].position[k] += beta * (swarm[j].position[k] - swarm[i].position[k]);
                swarm[i].position[k] += a * (rnd.NextDouble() - 0.5);
                if (swarm[i].position[k] < minX) swarm[i].position[k] = (maxX - minX) * rnd.NextDouble() + minX;
                if (swarm[i].position[k] > maxX) swarm[i].position[k] = (maxX - minX) * rnd.NextDouble() + minX;
              }
              swarm[i].error = Error(swarm[i].position);
              swarm[i].intensity = 1 / (swarm[i].error + 1); 
            } 
          } // j
        } // i each firefly

        Array.Sort(swarm); // low error to high
        if (swarm[0].error < bestError) // new best?
        {
          bestError = swarm[0].error;
          for (int k = 0; k < dim; ++k)
            bestPosition[k] = swarm[0].position[k];
        }
        ++epoch;
      } // while
      return bestPosition;
    } // Solve

    static double Distance(double[] posA, double[] posB)
    {
      double ssd = 0.0; // sum squared diffrences (Euclidean)
      for (int i = 0; i < posA.Length; ++i)
        ssd += (posA[i] - posB[i]) * (posA[i] - posB[i]);
      return Math.Sqrt(ssd);
    }

    static double Michalewicz(double[] xValues)
    {
      double result = 0.0;
      for (int i = 0; i < xValues.Length; ++i)
      {
        double a = Math.Sin(xValues[i]);
        double b = Math.Sin(((i+1) * xValues[i] * xValues[i]) / Math.PI);
        double c = Math.Pow(b, 20);
        result += a * c;
      }
      return -1.0 * result;
    } // Michalewicz

    static double Error(double[] xValues)
    {
      int dim = xValues.Length;
      double trueMin = 0.0;
      if (dim == 2)
        trueMin = -1.8013; // approx.
      else if (dim == 5)
        trueMin = -4.687658; // approx.
      else if (dim == 10)
        trueMin = -9.66015; // approx.
      double calculated = Michalewicz(xValues);
      return (trueMin - calculated) * (trueMin - calculated);
    }
  } // Program

  public class Firefly : IComparable<Firefly>
  {
    public double[] position;
    public double error;
    public double intensity; 

    public Firefly(int dim)
    {
      this.position = new double[dim];
      this.error = 0.0;
      this.intensity = 0.0;
    }

    public int CompareTo(Firefly other)
    {
      // allow auto sort low error to high
      if (this.error < other.error)
        return -1;
      else if (this.error > other.error)
        return +1;
      else
        return 0;
    }
  } // class Firefly
} // ns
